#MIT License
#Copyright (c) 2017 Tim Wentzlau

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" general DB handling in Kervi, store sensor values and settings """
import os
import time
from datetime import datetime
import json
import inspect
import threading
import sqlite3 as lite
import kervi.spine as spine
from kervi.core.utility.thread import KerviThread

SPINE = spine.Spine()
FILE_CON = None
MEMORY_CON = None
MEMORY_LOCK = threading.Lock()
FILE_LOCK = threading.Lock()

def create_file_con(db_name):
    return lite.connect(db_name + '.db', check_same_thread=False)

def create_memory_con(db_name):
    return lite.connect(db_name + "_mem.db", check_same_thread=False)
    #return lite.connect(":memory:", check_same_thread=False)


class MemoryCleanThread(KerviThread):
    def __init__(self):
        KerviThread.__init__(self)
        self.spine = spine.Spine()
        self.alive = False
        if self.spine:
            self.spine.register_command_handler("startThreads", self._start_command)
            self.spine.register_command_handler("stopThreads", self._stop_command)

    def _step(self):
        time.sleep(20)
        try:
            con = MEMORY_CON # create_memory_con()
            cursor = con.cursor()
            rows = cursor.execute("select count(*) from dynamicData")
            values = rows.fetchone()
            row_count = values[0]
            if row_count > 20000:
                limit = row_count - 20000
                cursor.execute("DELETE FROM dynamicData WHERE id IN (SELECT id FROM dynamicData ORDER BY id ASC LIMIT "+ str(limit) +");")
        except lite.Error as msg:
            SPINE.log.error("clean memory db, Command skipped: {0}, command{1}", msg)

    def _start_command(self):
        print("start storage service")
        if not self.alive:
            self.alive = True
            KerviThread.start(self)

    def _stop_command(self):
        if self.alive:
            self.alive = False
            self.stop()

DB_CREATE_SQL = """
        CREATE TABLE `log` (
            `id`	INTEGER PRIMARY KEY AUTOINCREMENT,
            `sourceId`	TEXT,
            `sourceName`	TEXT,
            `area`	TEXT,
            `data`	TEXT,
            `level`	INTEGER,
            `topic`	TEXT,
            `body`	TEXT,
            'ts'    `timeStamp`	REAL,
            `logType`	TEXT
        );
        CREATE TABLE "dynamicData" (
            `id`	INTEGER PRIMARY KEY AUTOINCREMENT,
            `dynamicValue`	TEXT,
            `value`	TEXT,
            `timeStamp`	TEXT
        );
        CREATE INDEX `sensorindex` ON `dynamicData` (`dynamicValue` ,`timeStamp` );
        CREATE TABLE "settings" (
            `id`	INTEGER PRIMARY KEY AUTOINCREMENT,
            `setting_group`	TEXT,
            `name`	TEXT,
            `value`	TEXT
        );
        CREATE INDEX `settingsindex` ON `settings` (`setting_group` ,`name` );
        CREATE TABLE "cron" (
            `id`	INTEGER PRIMARY KEY AUTOINCREMENT,
            `job_id` TEXT,
            `component_id` TEXT
            `name`	TEXT,
            `job_parameters` TEXT
        );
        CREATE TABLE "cron_meta" (
            `id`	INTEGER PRIMARY KEY AUTOINCREMENT,
            `cron_id`	INTEGER,
            `repeat_start` REAL NOT NULL,
            `repeat_end` REAL NOT NULL,
            `repeat_interval` varchar(255) NOT NULL,
            `repeat_year` varchar(255) NOT NULL,
            `repeat_month` varchar(255) NOT NULL,
            `repeat_day` varchar(255) NOT NULL,
            `repeat_week` varchar(255) NOT NULL,
            `repeat_weekday` varchar(255) NOT NULL
        );
        """


class _ObjectEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, "to_json"):
            return self.default(obj.to_json())
        elif hasattr(obj, "__dict__"):
            data = dict(
                (key, value)
                for key, value in inspect.getmembers(obj)
                if not key.startswith("__")
                and not inspect.isabstract(value)
                and not inspect.isbuiltin(value)
                and not inspect.isfunction(value)
                and not inspect.isgenerator(value)
                and not inspect.isgeneratorfunction(value)
                and not inspect.ismethod(value)
                and not inspect.ismethoddescriptor(value)
                and not inspect.isroutine(value)
            )
            return self.default(data)
        return obj


def execute_sql(con, sql):
    sql_commands = sql.split(';')

    for command in sql_commands:
        try:
            con.execute(command)
        except lite.Error as msg:
            SPINE.log.error("create db, Command skipped: {0}, command{1}", msg, command)


def init_db(db_name):
    global FILE_CON, MEMORY_CON

    MEMORY_CON = create_memory_con(db_name)
    FILE_CON = create_file_con(db_name)

    for con in [FILE_CON, MEMORY_CON]:
        cursor = con.cursor()
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        tables_exists = False
        tables = cursor.fetchall()
        for table in tables:
            tables_exists = True
            break

        if not tables_exists:
            execute_sql(con, DB_CREATE_SQL)        

    MemoryCleanThread()

#init_db()
TS_START = datetime.utcnow()

def store_dynamic_value(value_id, value, persist=False):
    try:
        if not persist:
            MEMORY_LOCK.acquire()
            cur = MEMORY_CON.cursor()
        else:
            FILE_LOCK.acquire()
            cur = FILE_CON.cursor()
        json_value = json.dumps(value["value"], cls=_ObjectEncoder, ensure_ascii=False).encode('utf8')
        cur.execute(
            "INSERT INTO dynamicData ('dynamicValue','value','timeStamp')  VALUES (?, ?, ?)",
            (value["id"], json_value, value["timestamp"])
        )
        if not persist:

            cur = MEMORY_CON.commit()
        else:
            cur = FILE_CON.commit()

    except lite.Error as er:
        SPINE.log.error('error store dynamic data, persist{0} error:{1}', persist, er)
    
    finally:
        if persist:
            FILE_LOCK.release()
        else:
            MEMORY_LOCK.release()

def store_setting(group, name, value):
    setting = retrieve_setting_db(group, name)
    FILE_LOCK.acquire()
    try:
        json_value = json.dumps(value, cls=_ObjectEncoder, ensure_ascii=False).encode('utf8')
        if setting:
            cur = FILE_CON.cursor()
            cur.execute(
                "update settings set value=? where id=?",
                (json_value, setting["id"])
            )
            FILE_CON.commit()
        else:
            cur = FILE_CON.cursor()
            cur.execute(
                "INSERT INTO settings ('setting_group','name','value')  VALUES (?, ?, ?)",
                (group, name, json_value)
            )
            FILE_CON.commit()

    except lite.Error as er:
        SPINE.log.error('error store settings data:{0}', er)
        print("error stire setting", er)
    except Exception as er:
        print("error stire setting", er)
    finally:
        FILE_LOCK.release()

def retrieve_setting_db(group, name):
    FILE_LOCK.acquire()
    try:
        cur = FILE_CON.cursor()
        cur.execute(
            "select * from settings where setting_group=? and name=?",
            (group, name)
        )

        all_rows = cur.fetchall()
        if len(all_rows) > 0:
            value = None
            try:
                value = json.loads(all_rows[0][3])
            except:
                SPINE.log.debug("retrieve setting failed:{0}/{1}", group, name)

            return {
                "id": all_rows[0][0],
                "group": all_rows[0][1],
                "name": all_rows[0][2],
                "value": value
            }
    finally:
        FILE_LOCK.release()

    return None

def retrieve_setting(group, name):
    setting = retrieve_setting_db(group, name)
    if setting:
        return setting["value"]

def get_dynamic_data(dynamic_value, date_from=None, date_to=None, limit=60):
    #print("gdf", date_from)
    #print("gdt", date_to)
    SPINE.log.debug("get dynamic data sensor:{0}, from {1} to {2}",dynamic_value, date_from, date_to)
    if date_from is None:
        date_from = TS_START
    elif isinstance(date_from, str):
        date_from = datetime.strptime(date_from,'%Y-%m-%dT%H:%M:%S.%fZ')


    if date_to is None:
        date_to = datetime.utcnow()
    elif isinstance(date_to, str):
        date_to = datetime.strptime(date_to,'%Y-%m-%dT%H:%M:%S.%fZ')

    #date_from_ts = (date_from - datetime(1970, 1, 1)).total_seconds()
    #date_to_ts = (date_to - datetime(1970, 1, 1)).total_seconds()

    try:
        con_list = [MEMORY_CON, FILE_CON]   
        result = []
        for con in con_list:
            if con == MEMORY_CON:
                MEMORY_LOCK.acquire()
            else:
                FILE_LOCK.acquire()
            try:
                cur = con.cursor()
                cur.execute(
                    "select * from dynamicData where dynamicValue=? and timestamp >= Datetime(?) and timestamp < Datetime(?)",
                    (dynamic_value, date_from, date_to)
                )

                all_rows = cur.fetchall()
                for row in all_rows:
                    result += [
                        {
                            "value":json.loads(row[2]),
                            "ts": row[3]
                        }
                    ]
            finally:
                if con == MEMORY_CON:
                    MEMORY_LOCK.release()
                else:
                    FILE_LOCK.release()
                if len(result):
                    break
        return result
    except lite.Error as er:
        SPINE.log.error('error get dynamic data:{0}', er)

def store_log_item(id, item):
    try:
        if not item["persist"]:
            MEMORY_LOCK.acquire()
            con = MEMORY_CON
        else:
            FILE_LOCK.acquire()
            con = FILE_CON

        cur = con.cursor()
        cur.execute(
            "INSERT INTO log ('sourceId','sourceName','area', 'data', 'level', 'topic', 'body', 'ts', 'logType')  VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
            (item["source_id"], item["source_name"], item["area"], item["data"], item["level"], item["topic"], item["body"], item["timestamp"], item["type"])
        )
        con.commit()
    except lite.Error as er:
        SPINE.log.error('error store log data:{0}', er)
    finally:
        if not item["persist"]:
            MEMORY_LOCK.release()
        else:
            FILE_LOCK.release()

def get_log_sort_field(item):
    return item["timestamp"]

def get_log_items(page, page_size, filters=None):
    connections = [FILE_CON, MEMORY_CON]
    result = []
    for con in connections:
        if con == MEMORY_CON:
            MEMORY_LOCK.acquire()
        else:
            FILE_LOCK.acquire()

        try:
            cur = con.cursor()
            cur.execute(
                "select * from log order by ts desc limit ?, ?",
                (page_size*page, page_size)
            )
            all_rows = cur.fetchall()
            for row in all_rows:
                if row[8]:
                    ts = datetime.fromtimestamp(row[8]).strftime('%Y-%m-%dT%H:%M:%SZ')
                else:
                    ts = None
                result += [
                    {
                        "source_id": row[1],
                        "source_name": row[2],
                        "area": row[3],
                        "data": row[4],
                        "level": row[5],
                        "topic": row[6],
                        "body": row[7],
                        "timestamp": row[8],
                        "type": row[9]
                    }
                ]
        finally:
            if con == MEMORY_CON:
                MEMORY_LOCK.release()
            else:
                FILE_LOCK.release()

    #result = sorted(result, get_log_sort_field)
    return result


def create_cron_job(component_id, job_id, name, job_parameters, start, end, repeat_interval, year,month):
    pass

def delete_cron_job():
    pass

def query_cron_job():
    pass

if (SPINE):
    SPINE.register_query_handler("getSensorData", get_dynamic_data)
    SPINE.register_command_handler("storeSetting", store_setting)
    SPINE.register_query_handler("retrieveSetting", retrieve_setting)
    SPINE.register_query_handler("getLogItems", get_log_items)
    SPINE.register_event_handler("valueChanged", store_dynamic_value)
    SPINE.register_event_handler("userLogMessage", store_log_item)
    SPINE.register_command_handler("createCronJob", create_cron_job)
    SPINE.register_command_handler("deleteCronJob", delete_cron_job)
    SPINE.register_query_handler("queryCronJob", query_cron_job)



#MIT License
#Copyright (c) 2017 Tim Wentzlau

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" Module that handles IPC between python processes  """
import uuid
from multiprocessing.connection import Listener, Client
import time
import threading
import kervi.spine as spine
#import sys
#import traceback
import kervi.utility.nethelper as nethelper

class _ConnCommandHandler(object):
    def __init__(self, command, conn, src):
        self.conn = conn
        self.command = command
        self.spine = spine.Spine()
        self.src = src
        self.spine.register_command_handler(command, self.on_command, injected="processSpine")

    def on_command(self, *args, **kwargs):
        try:
            injected = kwargs.get("injected", "")
            scope = kwargs.get("scope", "global")
            session = kwargs.get("session", None)
            if not injected == "processSpine":
                self.conn.send({"messageType":"command", "command":self.command, "args":args, "session":session})
        except IOError:
            self.spine.log.debug("IOError ConnCommandHandler:{0}", self.src)
        except:
            self.spine.log.exception("ConnCommandHandler")

class _ConnQueryHandler(object):
    id_count = 0
    uuid_handler = uuid.uuid4().hex
    def __init__(self, query, conn, process_spine, src):
        self.conn = conn
        self.query = query
        self.process_spine = process_spine
        self.spine = spine.Spine()
        self.src = src
        self.spine.register_query_handler(query, self.on_query, injected="processSpine")

    def on_query(self, *args, **kwargs):
        try:
            injected = kwargs.get("injected", "")
            scope = kwargs.get("scope", "global")
            session = kwargs.get("session", "session")
            if not injected == "processSpine":
                self.id_count += 1
                query_id = self.uuid_handler + "-" + str(self.id_count)
                event = threading.Event()
                event_data = {
                    "id":query_id,
                    "eventSignal":event,
                    "response":None,
                    "processed":False
                }
                self.process_spine.add_response_event(event_data)
                
                self.conn.send(
                    {"id":query_id, "messageType":"query", "query":self.query, "args":args, "session":session}
                )
                event.wait()
                res = event_data["response"]
                event_data["processed"] = True
                return res
        except IOError:
            self.spine.log.debug("IOError in ConnQueryHandler:{0}", self.src)
        except:
            self.spine.log.exception("ConnQueryHandler")

class _ConnEventHandler(object):
    def __init__(self, event, id_event, conn, src):
        self.conn = conn
        self.event = event
        self.id_event = id_event
        self.src = src
        self.spine = spine.Spine()
        self.spine.register_event_handler(event, self.on_event, id_event, injected="processSpine")

    def on_event(self, id_event, *args, **kwargs):
        try:
            injected = kwargs.get("injected", "")
            scope = kwargs.get("scope", "global")
            if not injected == "processSpine":
                self.spine.log.debug("trigger event: {0} on:{1} ", self.event, self.src)
                self.conn.send(
                    {"messageType":"event", "event":self.event, "id":id_event, "args":args}
                )
        except IOError:
            self.spine.log.debug("IOError ConnEventHandler:", self.src)
        except:
            self.spine.log.exception("ConnEventHandler")
            pass

class _ClientConnectionThread(threading.Thread):
    def __init__(self, process_spine):
        threading.Thread.__init__(self, None, None, "process spine clientconnection")
        self.daemon = True
        self.process_spine = process_spine
        self.terminate = False
    def run(self):
        try:
            self.process_spine.spine.log.debug("Listen on: {0}", self.process_spine.address)
            while not self.terminate:
                conn = self.process_spine.listener.accept()
                self.process_spine.spine.log.debug(
                    'connection accepted from {0}',
                    self.process_spine.listener.last_accepted
                )
                self.process_spine.add_process_connection(
                    conn,
                    self.process_spine.listener.last_accepted
                )
                if self.process_spine.is_root:
                    conn.send(
                        {"messageType":"processList", "list":self.process_spine.get_ipc_processes()}
                    )
        except IOError:
            self.process_spine.spine.log.debug("IOError in ClientConnectionThread")
        except:
            self.process_spine.spine.log.exception("ClientConnectionThread")

class _ConnectionMessageThread(threading.Thread):
    def __init__(self, process_spine, conn, src, is_root_connection=False):
        threading.Thread.__init__(self, None, None, "ConnectionMessage")
        self.daemon = True
        self.process_spine = process_spine
        self.connection = conn
        self.terminate = False
        self.is_root_connection = is_root_connection
        self.src = src

    def run(self):
        try:
            while not self.terminate:
                message = self.connection.recv()
                self.process_spine.handle_message(message, self.connection, self.src)
                time.sleep(0.001)
            print("connection closed:", self.src)
            self.connection.close()
        except EOFError:
            self.process_spine.spine.log.debug("connection messageThread eof from:{0}", self.src)
            self.process_spine.connection_terminated(self)
        except IOError:
            self.process_spine.spine.log.debug(
                "connection messageThread disconnected from:{0}",
                self.src
            )
            self.process_spine.connection_terminated(self)
        except:
            self.process_spine.spine.log.exception("connectionMessageThread", self.src)

class _ProcessSpine(object):
    def __init__(self, port, settings, **kwargs):
        self.is_root = kwargs.get("is_root", False)
        self.spine = spine.Spine()
        self.settings = settings
        self.port = port
        self.address = (self.settings["network"]["IPAddress"], port)
        #print("listen to:",self.address,self.settings["network"]["IPCSecret"])
        self.listener = Listener(self.address, authkey=self.settings["network"]["IPCSecret"])
        self.response_list = []
        self.process_list = []
        self.process_connections = []
        self.handlers = {"command":[], "query":[], "event":[]}


        self.spine.add_linked_spine(self)
        self.client_connection_thread = _ClientConnectionThread(self)
        self.client_connection_thread.start()
        self.register_at_root()


    def get_ipc_processes(self, *args, **kwargs):
        return self.process_list

    def register_at_root(self):
        if self.is_root:
            return

        root_address = (
            self.settings["network"]["IPRootAddress"],
            self.settings["network"]["IPCRootPort"]
        )

        conn = None
        reconnect = False
        while conn is None:
            try:
                if reconnect:
                    print("Try to connect to root:")
                conn = Client(
                    root_address,
                    authkey=self.settings["network"]["IPCSecret"]
                )
            except IOError:
                self.spine.log.debug("root not found")
                reconnect = True
                print("root not found", self.address)
                time.sleep(1)
        if reconnect:
            print("root found and connected")
        conn.send({"messageType":"registerProcess", "address":self.address})
        self.add_process_connection(conn, root_address, True)

    def register_process(self, process):
        self.process_list += [process]

    def add_process_connections(self, process_list):
        for process in process_list:
            if process != self.address:
                try:
                    conn = Client(process, authkey=self.settings["network"]["IPCSecret"])
                    self.add_process_connection(conn, process)
                    print("ok connection to:", process, "from:", self.address)
                except:
                    self.spine.log.exception("error add_process_connection")
                    print("error connection to:", process, "from:", self.address)

    def connection_terminated(self, connection):
        try:
            self.process_connections.remove(connection)
        except:
            pass

        for a in self.handlers:
            for h in self.handlers[a]:
                if h.src == connection.src:
                    try:
                        h.stop()
                        self.handlers[a].remove(connection)
                        break
                    except:
                        pass

        if connection.is_root_connection:
            self.register_at_root()

    def send_message(self, port, message):
        print ("SM")
        #self.client_connections[port].send(message)

    def add_process_connection(self, conn, src, is_root_connection=False):
        conn_thread = _ConnectionMessageThread(self, conn, src, is_root_connection)
        self.process_connections += [conn_thread]
        conn_thread.start()

        self.spine.log.debug("send rq to {1} clist:{0}", self.spine.get_commands(), src)
        for command in self.spine.get_commands():
            injected_count = 0
            handlers = self.spine.get_command_handlers(command)

            for handler in handlers:
                self.spine.log.debug("send rq handlers:{0}", handler)
                for ijc in self.handlers["command"]:
                    if ijc.on_command == handler:
                        injected_count += 1

            if injected_count < len(handlers):
                conn.send({"messageType":"registerCommandHandler", "command":command})

        self.spine.log.debug("send rq to {1} qlist:{0}", self.spine.get_queries(), src)
        for query in self.spine.get_queries():
            injected_count = 0
            handlers = self.spine.get_query_handlers(query)

            for handler in handlers:
                self.spine.log.debug("send rq handlers:{0}",handler)
                for ijq in self.handlers["query"]:
                    if ijq.on_query == handler:
                        injected_count += 1
            
            if injected_count < len(handlers):
                conn.send({"messageType":"registerQueryHandler", "query":query})

        # injected_events = []
        # for ije in self.handlers["event"]:
        #     injected_events += [{"id":ije.id_event, "event":ije.event, "matched":False, "src": ije.src}]

        # for event in self.spine.get_events():
        #     path = event.split("/")
        #     event_name = path[0]
        #     event_id = None
        #     if len(path) > 1:
        #         event_id = path[1]
        #     found = False
        #     for ie in injected_events:
        #         if ie["id"] == event_id and ie["event"] == event_name and not ie["matched"]:
        #             found = True
        #             ie["matched"] = True

        #     if not found:
        #         conn.send(
        #             {"messageType":"registerEventHandler", "eventId":event_id, "event":event_name}
        #         )
        self.spine.log.debug("send rq to {1} elist:{0}", self.spine.get_events(), src)
        for event in self.spine.get_events():
            path = event.split("/")
            event_name = path[0]
            event_id = None
            if len(path) > 1:
                event_id = path[1]

            injected_count = 0
            handlers = self.spine.get_event_handlers(event_name, event_id)

            for handler in handlers:
                self.spine.log.debug("send rq handlers:{0}",handler)
                for ije in self.handlers["event"]:
                    if ije.on_event == handler:
                        injected_count += 1

            if injected_count < len(handlers):
                conn.send({"messageType":"registerEventHandler", "eventId":event_id, "event":event_name})

    def close_all_connections(self):
        self.listener.close()
        for conn_thread in self.process_connections:
            conn_thread.terminate = True
            #conn_thread.join()
        self.client_connection_thread.terminate = True
        self.client_connection_thread.join()
        #time.sleep(1)

    def add_command_handler(self, command, connection, src):
        found = False
        for ch in self.handlers["command"]:
            if ch.command == command and ch.conn == connection:
                found = True
        if not found:
            self.handlers["command"] += [_ConnCommandHandler(command, connection, src)]

    def add_linked_command_handler(self, name, **kwargs):
        injected = kwargs.get("injected", "")
        if not injected == "processSpine":
            for conn_thread in self.process_connections:
                conn_thread.connection.send(
                    {"messageType":"registerCommandHandler", "command":name}
                )

    def add_query_handler(self, query, connection, src):
        found = False
        for qh in self.handlers["query"]:
            if qh.query == query and qh.conn == connection:
                found = True
        if not found:
            self.handlers["query"] += [_ConnQueryHandler(query, connection, self, src)]

    def add_linked_query_handler(self, name, **kwargs):
        injected = kwargs.get("injected", "")
        if not injected == "processSpine":
            for conn_thread in self.process_connections:
                conn_thread.connection.send({"messageType":"registerQueryHandler", "query":name})

    def add_linked_event_handler(self, name, id_event, **kwargs):
        injected = kwargs.get("injected", "")
        if not injected == "processSpine":
            for conn_thread in self.process_connections:
                conn_thread.connection.send(
                    {"messageType":"registerEventHandler", "event":name, "eventId":id_event}
                )

    def add_event_handler(self, event, id_event, connection, src):
        found = False
        for eh in self.handlers["event"]:
            if eh.event == event and eh.id_event == id_event and eh.conn == connection:
                found = True
        if not found:
            self.handlers["event"] += [_ConnEventHandler(event, id_event, connection, src)]

    def add_response_event(self, event):
        self.response_list += [event]

    def resolve_response(self, message):
        for event in self.response_list:
            if event["id"] == message["id"]:
                event["response"] = message["response"]
                event["eventSignal"].set()

    def handle_message(self, message, connection, src):
        try:
            self.spine.log.debug("message from:{0} on:{1} message:{2}", src, self.port, message)
            if message["messageType"] == "query":
                #print("query:", src, self.port,message["query"])
            
                res = self.spine.send_query(
                    message["query"],
                    *message["args"],
                    injected="processSpine",
                    session=message["session"]
                )
                #print("query:result", res)
                connection.send({"messageType":"queryResponse", "id":message["id"], "response":res})

            elif message["messageType"] == "queryResponse":
                self.resolve_response(message)
            elif message["messageType"] == "registerQueryHandler":
                self.add_query_handler(message["query"], connection, src)
            elif message["messageType"] == "command":
                self.spine.send_command(
                    message["command"],
                    *message["args"],
                    injected="processSpine",
                    session=message["session"]
                )
            elif message["messageType"] == "registerCommandHandler":
                self.add_command_handler(message["command"], connection, src)
            elif message["messageType"] == "event":
                self.spine.trigger_event(
                    message["event"],
                    message["id"],
                    *message["args"],
                    injected="processSpine"
                )
            elif message["messageType"] == "registerEventHandler":
                self.add_event_handler(message["event"], message["eventId"], connection, src)
            elif message["messageType"] == "processList":
                self.add_process_connections(message["list"])
            elif message["messageType"] == "registerProcess":
                self.register_process(message["address"])
        except:
            self.spine.log.exception("on process message")
